// ----------------------------------------------------------------------------
// (C) 2004
//     carrara engineering GmbH
//     Switzerland
//     http://www.carrara.ch
//     mailto:info@carrara.ch
// ----------------------------------------------------------------------------
// Permission to use, copy, modify, distribute and sell this software for any 
//     purpose is hereby granted without fee, provided that the above copyright 
//     notice appear in all copies and that both that copyright notice and this 
//     permission notice appear in supporting documentation.
// The author make no representations about the suitability 
//     of this software for any purpose. It is provided "as is" 
//     without express or implied warranty.
// ----------------------------------------------------------------------------


#include <iostream>
#include <windows.h>

using std::cout;
using std::endl;

// ----------------------------------------------------------------------------

namespace Consensus1
{

bool testAndSet(volatile bool& i_raFlag)
{
  // Start Atomic
  bool bFlag = i_raFlag;
  i_raFlag   = true;
  return bFlag;
  // End Atomic
}

typedef int T_Type;

volatile T_Type s_aPreference[2];
volatile bool   s_bSync = false;

T_Type getP1Consensus(const T_Type& i_raValue)
{
  s_aPreference[0] = i_raValue;
  if (testAndSet(s_bSync)==false)
  {
    return s_aPreference[0];
  }
  else
  {
    return s_aPreference[1];
  }
}

T_Type getP2Consensus(const T_Type& i_raValue)
{
  s_aPreference[1] = i_raValue;
  if (testAndSet(s_bSync)==false)
  {
    return s_aPreference[1];
  }
  else
  {
    return s_aPreference[0];
  }
}

void test()
{
  s_bSync = false;
  
  T_Type nResult1 = getP1Consensus(1);
  T_Type nResult2 = getP2Consensus(2);
  cout << nResult1 << ":" << nResult2 << endl;

  s_bSync = false;
  
  nResult2 = getP1Consensus(2);
  nResult1 = getP2Consensus(1);
  cout << nResult1 << ":" << nResult2 << endl;
}

}
// ----------------------------------------------------------------------------

namespace Consensus2
{

bool testAndSet(volatile long& i_raFlag)
{
  return InterlockedExchange(&i_raFlag,true)?true:false;
}

typedef int T_Type;

volatile T_Type s_aPreference[2];
volatile long   s_bSync = false;

T_Type getP1Consensus(const T_Type& i_raValue)
{
  s_aPreference[0] = i_raValue;
  if (testAndSet(s_bSync)==false)
  {
    return s_aPreference[0];
  }
  else
  {
    return s_aPreference[1];
  }
}

T_Type getP2Consensus(const T_Type& i_raValue)
{
  s_aPreference[1] = i_raValue;
  if (testAndSet(s_bSync)==false)
  {
    return s_aPreference[1];
  }
  else
  {
    return s_aPreference[0];
  }
}

void test()
{
  s_bSync = false;
  
  T_Type nResult1 = getP1Consensus(1);
  T_Type nResult2 = getP2Consensus(2);
  cout << nResult1 << ":" << nResult2 << endl;

  s_bSync = false;
  
  nResult2 = getP1Consensus(2);
  nResult1 = getP2Consensus(1);
  cout << nResult1 << ":" << nResult2 << endl;
}

}

// ----------------------------------------------------------------------------

namespace Consensus3
{

bool testAndSet(volatile long& i_raFlag)
{
  return InterlockedExchange(&i_raFlag,true)?true:false;
}

typedef int   T_Type;

// Global State
volatile T_Type s_aBuffer[3];
volatile long   s_bSync;
volatile bool   s_bReadSignal;
volatile int    s_nLastWritten;
volatile int    s_nReaderPref;
volatile int    s_nWriterPref;

// Writers State
int s_nConsented;

int getReadConsensus()
{
  s_bSync       = false;
  s_bReadSignal = true;

  s_nReaderPref = s_nLastWritten;
  if (testAndSet(s_bSync)==false)
  {
    return s_nReaderPref;
  }
  else
  {
    return s_nWriterPref;
  }
}

int getWriteConsensus()
{
  if (s_bReadSignal)
  {
    s_bReadSignal = false;
    
    s_nWriterPref = s_nLastWritten;
    if (testAndSet(s_bSync)==false)
    {
      s_nConsented = s_nWriterPref;
    }
    else
    {
      s_nConsented = s_nReaderPref;
    }
  }
  return s_nConsented;
}

void init(int i_nDefault)
{
  s_aBuffer[0]  = i_nDefault;
  s_bSync       = false;
  s_bReadSignal = true;
  s_nLastWritten  = 0;
}

void testReader()
{
   int nConsented = getReadConsensus();
   cout << "Read[" << nConsented << "]: " << s_aBuffer[nConsented] << endl;
}

void testWriter(int i_nData)
{
   static const int nPermutator[3][3] 
          = {{1,2,1}
           ,{2,2,0}
           ,{1,0,0}};

   int nConsented   = getWriteConsensus();
   int nSafe        = nPermutator[nConsented][s_nLastWritten];
   s_aBuffer[nSafe] = i_nData;
   s_nLastWritten   = nSafe;
}

void test()
{
  init(99);
  testReader();
  testWriter(11);
  testReader();
  testReader();
  testWriter(22);
  testReader();
  testWriter(33);
  testReader();
  testWriter(44);
  testReader();
}

}

// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------

namespace Consensus4
{

bool testAndSet(volatile long& i_raFlag)
{
  return InterlockedExchange(&i_raFlag,true)?true:false;
}

typedef int   T_Type;

// Global State
volatile T_Type s_aBuffer[3];
volatile long   s_bSync;
volatile bool   s_bReadSignal;
volatile int    s_nLastWritten;
volatile int    s_nReaderPref;
volatile int    s_nWriterPref;

// Writers State
int s_nConsented;

int getReadConsensus()
{
  s_bSync       = false;
  s_bReadSignal = true;

  register int nReaderPref = s_nLastWritten;
  s_nReaderPref = nReaderPref;
  register int nPref;
  if (testAndSet(s_bSync)==false)
  {
    nPref = s_nReaderPref;
  }
  else
  {
    nPref = s_nWriterPref;
  }
  return nPref;
}

int getWriteConsensus()
{
  register int nConsented;
  if (s_bReadSignal)
  {
    s_bReadSignal = false;
    
    register nWriterPref = s_nLastWritten;
    s_nWriterPref = nWriterPref;
    if (testAndSet(s_bSync)==false)
    {
      nConsented = s_nWriterPref;
    }
    else
    {
      nConsented = s_nReaderPref;
    }
    s_nConsented = nConsented;
  }
  else
  {
    nConsented = s_nConsented;
  }
  return nConsented;
}

void init(int i_nDefault)
{
  s_aBuffer[0]  = i_nDefault;
  s_bSync       = false;
  s_bReadSignal = true;
  s_nLastWritten  = 0;
}

void testReader()
{
   int nConsented = getReadConsensus();
   cout << "Read[" << nConsented << "]: " << s_aBuffer[nConsented] << endl;
}

void testWriter(int i_nData)
{
   static const int nPermutator[3][3] 
          = {{1,2,1}
           ,{2,2,0}
           ,{1,0,0}};

   int nConsented   = getWriteConsensus();
   int nSafe        = nPermutator[nConsented][s_nLastWritten];
   s_aBuffer[nSafe] = i_nData;
   s_nLastWritten   = nSafe;
}

void test()
{
  init(99);
  testReader();
  testWriter(11);
  testReader();
  testReader();
  testWriter(22);
  testReader();
  testWriter(33);
  testReader();
  testWriter(44);
  testReader();
}

}

// ----------------------------------------------------------------------------

// ----------------------------------------------------------------------------

namespace Consensus5
{

bool testAndSet(volatile long& i_rbFlag)
{
  return InterlockedExchange(&i_rbFlag,true)?true:false;
}

class SharedData
{
  public:
    SharedData()
    : m_bSync        (false)
    , m_bReadSignal  (true)
    , m_nLastWritten (0)
    , m_nReaderPref  (0)
    , m_nWriterPref  (0)
    {}

    volatile long m_bSync;
    volatile bool m_bReadSignal;
    volatile int  m_nLastWritten;
    volatile int  m_nReaderPref;
    volatile int  m_nWriterPref;
};

template <class T>
class Buffer
{
  public:
    Buffer(const T& i_raDefaultValue)
    {
      m_aBuffer[0] = i_raDefaultValue;
    }
    
    volatile T& operator[](int i_nIdx)
    {
      return m_aBuffer[i_nIdx];
    }
    
  private:
    volatile T m_aBuffer[3];
};

class ReadConsensus
{
  public:
    ReadConsensus(SharedData& i_raSharedData)
    : m_raSharedData (i_raSharedData)
    {}

    int getConsensus()
    {
      m_raSharedData.m_bSync       = false;
      m_raSharedData.m_bReadSignal = true;

      m_raSharedData.m_nReaderPref = m_raSharedData.m_nLastWritten;
      if (testAndSet(m_raSharedData.m_bSync)==false)
      {
        return m_raSharedData.m_nReaderPref;
      }
      else
      {
        return m_raSharedData.m_nWriterPref;
      }
    }

  private:
    SharedData& m_raSharedData;
};

class WriteConsensus
{
  public:
    WriteConsensus(SharedData& i_raSharedData)
    : m_raSharedData (i_raSharedData)
    , m_nConsented (0)
    {}

    int getConsensus()
    {
      if (m_raSharedData.m_bReadSignal)
      {
        m_raSharedData.m_bReadSignal = false;

        m_raSharedData.m_nWriterPref = m_raSharedData.m_nLastWritten;
        if (testAndSet(m_raSharedData.m_bSync)==false)
        {
          m_nConsented = m_raSharedData.m_nWriterPref;
        }
        else
        {
          m_nConsented = m_raSharedData.m_nReaderPref;
        }
      }
      return m_nConsented;
    }
    
    int getLastWritten() const
    {
      return m_raSharedData.m_nLastWritten;
    }
    
    void setLastWritten(int i_nLastWritten) 
    {
      m_raSharedData.m_nLastWritten = i_nLastWritten;
    }

  private:
    SharedData& m_raSharedData;
    int         m_nConsented;
};

void testReader(ReadConsensus& i_raReadConsensus, Buffer<int>& i_raBuffer)
{
   int nEngaged = i_raReadConsensus.getConsensus();
   cout << "Read[" << nEngaged << "]: " << i_raBuffer[nEngaged] << endl;
}

void testWriter(WriteConsensus& i_raWriteConsensus, Buffer<int>& i_raBuffer, int i_nData)
{
   static const int nPermutator[3][3] = {{1,2,1}
                                       ,{2,2,0}
                                       ,{1,0,0}};

   int nLastEngaged      = i_raWriteConsensus.getConsensus();
   int nLastWritten      = i_raWriteConsensus.getLastWritten();
   int nWriteIdx         = nPermutator[nLastEngaged][nLastWritten];
   i_raBuffer[nWriteIdx] = i_nData;
   i_raWriteConsensus.setLastWritten(nWriteIdx);
}

void test()
{
   Buffer<int>     aBuffer(99);

   SharedData      aSharedData;
   ReadConsensus   aReader(aSharedData);
   WriteConsensus  aWriter(aSharedData);
   
   testReader(aReader,aBuffer);
   testWriter(aWriter,aBuffer,11);
   testReader(aReader,aBuffer);
   testReader(aReader,aBuffer);
   testWriter(aWriter,aBuffer,22);
   testReader(aReader,aBuffer);
   testWriter(aWriter,aBuffer,33);
   testReader(aReader,aBuffer);
   testWriter(aWriter,aBuffer,44);
   testReader(aReader,aBuffer);
}

}


// ----------------------------------------------------------------------------


main()
{
  cout << "--------------------------- Test 1 ---------------------------" << endl;
  Consensus1::test();

  cout << "--------------------------- Test 2 ---------------------------" << endl;
  Consensus2::test();

  cout << "--------------------------- Test 3 ---------------------------" << endl;
  Consensus3::test();

  cout << "--------------------------- Test 4 ---------------------------" << endl;
  Consensus4::test();

  cout << "--------------------------- Test 5 ---------------------------" << endl;
  Consensus5::test();

  return 0;
}
